<?php

declare(strict_types=1);

namespace ConnectPay\Payment\Gateway\Command;

use ConnectPay\Payment\Gateway\Config\Config;
use ConnectPay\Payment\Gateway\Http\Client\Guzzle;
use Exception;
use Magento\Framework\Serialize\Serializer\Json;
use Magento\Payment\Gateway\Command\CommandException;
use Magento\Payment\Gateway\Command\ResultInterface;
use Magento\Payment\Gateway\Command\ResultInterface as CommandResultInterface;
use Magento\Payment\Gateway\CommandInterface;
use Magento\Payment\Gateway\ErrorMapper\ErrorMessageMapperInterface;
use Magento\Payment\Gateway\Http\TransferFactoryInterface;
use Magento\Payment\Gateway\Request\BuilderInterface;
use Magento\Payment\Gateway\Response\HandlerInterface;
use Magento\Payment\Gateway\Validator\ValidatorInterface;
use Magento\Framework\Webapi\Rest\Request;
use Psr\Log\LoggerInterface as Logger;

class RequestCommand implements CommandInterface
{
    /**
     * @var BuilderInterface
     */
    private $requestBuilder;

    /**
     * @var TransferFactoryInterface
     */
    private $transferFactory;

    /**
     * @var Guzzle
     */
    private $client;

    /**
     * @var CommandResultInterface
     */
    private $result;

    /**
     * @var HandlerInterface
     */
    private $handler;

    /**
     * @var ValidatorInterface
     */
    private $validator;

    /**
     * @var ErrorMessageMapperInterface
     */
    private $errorMessageMapper;

    /**
     * @var Json
     */
    private $serializer;

    private $logger;

    public function __construct(
        BuilderInterface            $requestBuilder,
        TransferFactoryInterface    $transferFactory,
        Guzzle                      $client,
        CommandResultInterface      $result,
        Json                        $serializer,
        HandlerInterface            $handler = null,
        ValidatorInterface          $validator = null,
        ErrorMessageMapperInterface $errorMessageMapper = null
    )
    {
        $this->requestBuilder = $requestBuilder;
        $this->transferFactory = $transferFactory;
        $this->client = $client;
        $this->result = $result;
        $this->serializer = $serializer;
        $this->handler = $handler;
        $this->validator = $validator;
        $this->errorMessageMapper = $errorMessageMapper;
    }

    /**
     * @param array $commandSubject
     * @return ResultInterface|void|null
     * @throws Exception
     */
    public function execute(array $commandSubject): ?ResultInterface
    {
        $transferObject = $this->transferFactory->create(
            $this->requestBuilder->build($commandSubject)
        );
        $data = $transferObject->getBody();

        $response = $this->client->doRequest($data, Request::HTTP_METHOD_POST, Config::API_REQUEST_INITIATE_PAYMENT_ENDPOINT);
        $response = $response->getBody()->getContents();

        if (!empty($response)) {
            $responseArray = $this->serializer->unserialize($response);
        } else return null;

        if ($this->validator !== null) {
            $result = $this->validator->validate(array_merge($commandSubject, ['response' => $responseArray]));
            if (!$result->isValid()) {
                $this->processErrors($result);
            }
        }

        if ($this->handler) {
            $this->handler->handle(
                $commandSubject,
                $responseArray
            );
        }

        if ($this->result) {
            return $this->result->addData(
                array_merge($commandSubject, ['response' => $responseArray])
            );
        }
    }

    /**
     * Tries to map error messages from validation result and logs processed message.
     * Throws an exception with mapped message or default error.
     *
     * @param ResultInterface $result
     * @throws CommandException
     */
    private function processErrors(ResultInterface $result)
    {
        $messages = [];
        $errorsSource = array_merge($result->getErrorCodes(), $result->getFailsDescription());
        foreach ($errorsSource as $errorCodeOrMessage) {
            $errorCodeOrMessage = (string)$errorCodeOrMessage;

            // error messages mapper can be not configured if payment method doesn't have custom error messages.
            if ($this->errorMessageMapper !== null) {
                $mapped = (string)$this->errorMessageMapper->getMessage($errorCodeOrMessage);
                if (!empty($mapped)) {
                    $messages[] = $mapped;
                    $errorCodeOrMessage = $mapped;
                }
            }
            $this->logger->critical('Payment Error: ' . $errorCodeOrMessage);
        }

        throw new CommandException(
            !empty($messages)
                ? __(implode(PHP_EOL, $messages))
                : __('Transaction has been declined. Please try again later.')
        );
    }
}
